from typing import Iterable
from typing import Optional

from .filter_graph import Filter
from .filter_graph import FilterOption
from .filter_graph import Stream


def concat(graphs: Iterable[Stream], n: Optional[int] = None, v: Optional[int] = None, a: Optional[int] = None) -> Stream:
    """Concatenate audio and video streams, joining them together one after the other.

    :param int n: set the number of segments. Default is 2.
    :param int v: set the number of output video streams, that is also the number of video streams in each segment. Default is 1.
    :param int a: set the number of output audio streams, that is also the number of audio streams in each segment. Default is 0.
    """
    filter_ = Filter(
        command="concat",
        params=[
            FilterOption(name="n", value=n),
            FilterOption(name="v", value=v),
            FilterOption(name="a", value=a),
        ],
    )
    for graph in graphs:
        filter_.add_input(graph)
    return Stream().append(filter_)


def overlay(
    main: Stream,
    sub: Stream,
    x: Optional[str] = None,
    y: Optional[str] = None,
    eof_action: Optional[str] = None,
    eval: Optional[str] = None,
    shortest: Optional[bool] = None,
    format: Optional[str] = None,
    repeatlast: Optional[bool] = None,
    alpha: Optional[str] = None,
) -> Stream:
    """Overlay a video source on top of the input.

    :param str x: set the x expression
    :param str y: set the y expression
    :param str eof_action: Action to take when encountering EOF from secondary input
            possible values: repeat, endall, pass
    :param str eval: specify when to evaluate expressions
            possible values: init, frame
    :param bool shortest: force termination when the shortest input terminates
    :param str format: set output format
            possible values: yuv420, yuv420p10, yuv422, yuv422p10, yuv444, yuv444p10, rgb, gbrp, auto
    :param bool repeatlast: repeat overlay of the last overlay frame
    :param str alpha: alpha format
            possible values: straight, premultiplied
    """
    filter_ = Filter(
        command="overlay",
        filter_type="AVMEDIA_TYPE_VIDEO",
        params=[
            FilterOption(name="x", value=x),
            FilterOption(name="y", value=y),
            FilterOption(name="eof_action", value=eof_action),
            FilterOption(name="eval", value=eval),
            FilterOption(name="shortest", value=shortest),
            FilterOption(name="format", value=format),
            FilterOption(name="repeatlast", value=repeatlast),
            FilterOption(name="alpha", value=alpha),
        ],
    )
    filter_.add_input(main)
    filter_.add_input(sub)
    return Stream().append(filter_)



"""compatibility layer to match the kkroening library.
@TODO: better docstrings"""

from functools import singledispatch
from typing import List
from typing import Tuple

from zprp_ffmpeg2.base_connector import BaseConnector

from .filter_graph import Filter
from .filter_graph import FilterOption
from .filter_graph import MergeOutputFilter
from .filter_graph import SinkFilter
from .filter_graph import SourceFilter
from .filter_graph import Stream
from .process_connector import ProcessConnector


# this one is unique, creates the Stream object
def input(filename: str, **kwargs) -> Stream:
    source = SourceFilter(filename, **kwargs)
    return Stream().append(source)


def filter(stream_spec: Stream, filter_name: str, *args, **kwargs) -> Stream:
    """Applies a custom filter"""
    options = []
    for arg in args:
        options.append(FilterOption(arg, None))
    for name, value in kwargs.items():
        options.append(FilterOption(name, value))
    new_stream = stream_spec.append(Filter(command=filter_name, params=options))
    return new_stream


def output(stream: Stream, filename: str) -> Stream:
    sink = SinkFilter(filename)
    new_stream = stream.append(sink)
    return new_stream


def global_args(stream: Stream, *args) -> Stream:
    new_args: List[str] = []
    for arg in args:
        new_args.append(str(arg))
    stream.global_options = new_args
    return stream


@singledispatch
def get_args(arg):
    return "Unknown type"


@get_args.register(Stream)
def _(stream: Stream, overwrite_output: bool = False) -> List[str]:
    """Build command-line arguments to be passed to ffmpeg."""
    args = ProcessConnector.compile(stream).split()
    if overwrite_output:
        args += ["-y"]
    return args


@get_args.register(list)
def _(streams: list[Stream], overwrite_output: bool = False) -> List[str]:
    """Build command-line arguments to be passed to ffmpeg."""
    streams.reverse()
    new_stream = Stream().append(MergeOutputFilter(streams))
    args = ProcessConnector.compile(new_stream).split()
    if overwrite_output:
        args += ["-y"]
    return args


def compile(stream: Stream, cmd: str = "ffmpeg", overwrite_output: bool = False) -> List[str]:
    """Returns command-line for invoking ffmpeg split by space"""
    return [cmd, *get_args(stream, overwrite_output)]


# this api always uses process
def run(stream: Stream, extra_options: str = "") -> Tuple[bytes, bytes]:
    """Returns (stdout,stderr) tuple"""
    return ProcessConnector.run(stream, extra_options).communicate()


# this api always uses process
def run_async(stream: Stream) -> BaseConnector:
    """Returns handle to a process. Can raise an exception if script tries to terminate before ffmpeg is done."""
    return ProcessConnector.run(stream)


def overwrite_output(stream: Stream) -> Stream:
    stream.global_options.append("-y")
    return stream


def merge_outputs(*streams: Stream) -> Stream:
    """Include all given outputs in one ffmpeg command line"""
    return Stream().append(MergeOutputFilter(streams))



"""There is no general 'Node' class, because it doesn't work well with object relations, source and sink filters are kind of orthogonal.
It slightly violates DRY, but the parameter types are different. It is what it is"""

from collections import OrderedDict
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Union

from ordered_set import OrderedSet


class FilterType(Enum):
    VIDEO = "AVMEDIA_TYPE_VIDEO"
    AUDIO = "AVMEDIA_TYPE_AUDIO"

    @classmethod
    def to_command_string(cls, value: Any) -> str:
        return ":v]" if value == FilterType.VIDEO.value else ":a]" if value == FilterType.AUDIO.value else ""


@dataclass
class FilterOption:
    name: str
    value: Any


@dataclass
class ComplexCommand:
    command: str = ""
    file: Optional[str] = None
    params: str = ""
    filter_type: str = ""
    filter_type_command: str = ""
    kwargs: str = ""


@dataclass
class MergeOutputCommand(ComplexCommand):
    inputs: str = ""
    outputs: str = ""


class Filter:
    """Filters can have many inputs and many outputs, holds the filter name and potential params"""

    _filter_counter: ClassVar[Dict[str, int]] = defaultdict(int)

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        if "command" in kwargs:
            cls._filter_counter[kwargs["command"]] += 1
            instance._id = cls._filter_counter[kwargs["command"]]
        return instance

    def __init__(self, *, command: str, params: Optional[List[FilterOption]] = None, filter_type: str = FilterType.VIDEO.value):
        self._out: List[AnyNode] = []
        self._in: List[AnyNode | "Stream"] = []
        self.command = command
        self.params = params if params else []
        self.filter_type = filter_type

    def add_output(self, parent: "Filter | SinkFilter"):
        self._out.append(parent)

    def add_input(self, child: "Filter | SourceFilter | Stream"):
        self._in.append(child)

    def get_command(self):
        joined_params = ":".join(p.name + "=" + str(p.value) for p in self.params if p.value)
        if joined_params:  # if there was no option, leave empty string
            joined_params = "=" + joined_params
        filter_type_command = FilterType.to_command_string(self.filter_type)
        return ComplexCommand(
            command=self.command,
            params=joined_params,
            filter_type=self.filter_type,
            filter_type_command=filter_type_command,
        )


# names as per ffmpeg documentation
class SourceFilter:
    def __init__(self, in_path: str, **kwargs):
        self.in_path: str = in_path
        self._out: List[Union["Filter", "SinkFilter"]] = []
        self.kwargs = kwargs

    def add_output(self, parent: Union["Filter", "SinkFilter"]):
        self._out.append(parent)

    def add_input(self, child: "Filter"):
        raise NotImplementedError("This node can't have inputs")

    def get_command(self) -> ComplexCommand:
        return ComplexCommand(
            kwargs=convert_kwargs_to_cmd_args(self.kwargs),
            file=self.in_path,
        )


class SinkFilter:
    def __init__(self, out_path: str):
        self.out_path: str = out_path
        self._in: List[Union["Filter", "SourceFilter"]] = []
        self._out: List[Union["Filter", "MergeOutputFilter"]] = []

    def add_input(self, parent: Union["Filter", "SourceFilter"]):
        self._in.append(parent)

    def add_output(self, child: Union["Filter", "MergeOutputFilter"]):
        if isinstance(child, MergeOutputFilter):
            self._out.append(child)
        raise NotImplementedError("This node can't have outputs")

    def get_command(self) -> ComplexCommand:
        return ComplexCommand(
            file=self.out_path,
        )


# in python 3.12 there is 'type' keyword, but we are targetting 3.8
# https://stackoverflow.com/questions/76712720/typeerror-unsupported-operand-types-for-type-and-nonetype
# python >3.9 uses , instead of | idk if it works with python 3.12
AnyNode = Union[Filter, SourceFilter, SinkFilter]


class Stream:
    def __init__(self) -> None:
        self._nodes: List[Union[Filter, SourceFilter, SinkFilter]] = []
        self.global_options: List[str] = []

    def append(self, node: Union[Filter, SourceFilter, SinkFilter, "MergeOutputFilter"]) -> "Stream":
        # Create a deepcopy of the current instance
        new_stream = deepcopy(self)
        if len(new_stream._nodes) > 0:
            # Connect the last node to the new one
            if not isinstance(new_stream._nodes[-1], SinkFilter) and not isinstance(node, SourceFilter):
                new_stream._nodes[-1].add_output(node)
                node.add_input(new_stream._nodes[-1])
        # Append the new node
        new_stream._nodes.append(node)
        return new_stream  # Return the new instance

    def output(self, out_path: str) -> "Stream":
        # Add a SinkFilter and return a new Stream
        sink = SinkFilter(out_path)
        return self.append(sink)


class MergeOutputFilter:
    """Represents a filter that merges multiple streams into one.

    Attributes:
        streams (Iterable[Stream]): The streams to be merged.
        _in (List[Union[Filter, SourceFilter, SinkFilter]]): List of input nodes connected to this merge filter.
    """

    def __init__(self, streams: Iterable[Stream]):
        self.streams = streams
        self._in: List[Union["Filter", "SourceFilter", "SinkFilter"]] = []

    def add_output(self, parent: "Filter"):
        raise NotImplementedError("This node can't have outputs")

    def add_input(self, child: "Filter"):
        raise NotImplementedError("This node can't have inputs")

    def get_command(self) -> ComplexCommand:
        """Method only made for compatibility with other nodes in FilterParser.
        Returns empty command, which is not used in practice.
        """
        return MergeOutputCommand()


class FilterParser:
    def __init__(self):
        self.multi_input = ["concat", "overlay"]
        self.inputs_counter = 0
        self.outputs_counter = 0
        self.filter_counter = 0
        self.result_counter = 0
        self.merge_counter = 0

        self.inputs = []
        self.outputs = []
        self.filters = []
        self.input_file_to_index = OrderedDict()

    def generate_command(self, stream: Stream, last=None) -> str:  # type: ignore
        for node in stream._nodes:
            command_obj = node.get_command()
            command = command_obj.command
            filter_type_command = command_obj.filter_type_command
            params = command_obj.params
            map_cmd = "-map"
            i_cmd = "-i"
            file = command_obj.file

            # many inputs one output
            if any(filter_ in command for filter_ in self.multi_input):
                last_results = []
                for graph in node._in:  # type: ignore
                    last_results.append(self.generate_command(graph))  # type: ignore
                results = "".join([f"[{result}]" for result in last_results])
                self.filters.append(f"{results}{command}{params}[v{self.result_counter}];")
                last = f"v{self.result_counter}"
                self.result_counter += 1
            # input
            elif isinstance(node, SourceFilter):
                kwargs = command_obj.kwargs
                if file not in self.input_file_to_index:
                    self.input_file_to_index[file] = len(self.input_file_to_index)
                    self.inputs.append(f"{kwargs} {i_cmd} {file}")
                last = self.input_file_to_index[file]
                continue
            # output
            elif isinstance(node, SinkFilter):
                if last == 0 or (len(self.input_file_to_index) == 1 and isinstance(last, int)):
                    self.outputs.append(f"{file}")
                elif isinstance(last, int) and len(self.input_file_to_index) > 1:
                    self.outputs.append(f"{map_cmd} {last} {file}")
                elif len(self.input_file_to_index) > 1 or isinstance(last, str):
                    self.outputs.append(f"{map_cmd} [{last}] {file}")
                self.outputs_counter += 1
                continue
            # single input single output
            # merge output
            elif isinstance(node, MergeOutputFilter):
                merge_last = None
                for sub_stream in node.streams:
                    merge_last = self.generate_command(sub_stream, merge_last)
                    last = merge_last
                    self.outputs_counter = 0
                    self.filter_counter = 0
                    self.result_counter = 0
                self.merge_counter += 1
                continue
            else:
                if isinstance(last, int):
                    self.filters.append(f"[{last}{filter_type_command}{command}{params}[v{self.result_counter}];")
                else:
                    self.filters.append(f"[{last}]{command}{params}[v{self.result_counter}];")
                last = f"v{self.result_counter}"
                self.result_counter += 1

        if len(self.filters) == 0 and not all(isinstance(node, (SourceFilter, SinkFilter)) for node in stream._nodes):
            return ""
        return last

    def generate_result(self, stream: Stream) -> str:
        self.generate_command(stream)
        self.filters = list(OrderedSet(self.filters))

        if self.merge_counter == 1:
            for node in stream._nodes:
                if isinstance(node, MergeOutputFilter):
                    break
            if len(self.filters) == 0:
                return " ".join(list(OrderedSet(self.inputs))) + " " + " ".join(self.outputs) + " " + " ".join(stream.global_options)

            filters = "".join(self.filters)[:-1]
            o_counter = 0
            new_outputs = []
            mapped_outs = [out for out in self.outputs if "-map" in out]
            other_outs = [out for out in self.outputs if out not in mapped_outs]
            for output in mapped_outs[:-1]:
                k = output.find("[")
                l = output.find("]")
                out = output[k + 1 : l]
                j = filters.find(out)
                new_out = f"[out{o_counter}]"
                filters = filters[: j - 1] + ",split" + new_out + filters[j - 1 :]
                new_output = output[:k] + new_out + output[l + 1 :]
                new_outputs.append(new_output)
                o_counter += 1
            last_output = mapped_outs[-1]
            self.outputs = new_outputs
            self.outputs.append(last_output)
            self.outputs.extend(other_outs)

            return (
                " ".join(list(OrderedSet(self.inputs)))
                + ' -filter_complex "'
                + filters
                + '" '
                + " ".join(self.outputs)
                + " "
                + " ".join(stream.global_options)
            )

        elif len(self.filters) == 0:
            return " ".join(self.inputs) + " " + self.outputs[-1].split()[-1] + " " + " ".join(stream.global_options)

        return (
            " ".join(self.inputs)
            + ' -filter_complex "'
            + " ".join(self.filters)[:-1]
            + '" '
            + " ".join(self.outputs)
            + " "
            + " ".join(stream.global_options)
        )


def convert_kwargs_to_cmd_args(kwargs: Dict[str, Any]) -> str:
    """Converts keyword arguments to FFmpeg command-line arguments.

    Args:
        kwargs (Dict[str, Any]): Dictionary of keyword arguments.

    Returns:
        str: A string of command-line arguments suitable for FFmpeg.
    """
    args = []
    for k, v in kwargs.items():
        if isinstance(v, bool):
            prefix = "" if v else "no"
            args.append(f"-{prefix}{k}")
            continue
        args.append(f"-{k}")
        args.append(str(v))
    return " ".join(args)



__version__ = "2.2.0"
# from . import longest
from inspect import getmembers
from inspect import isfunction
from typing import Any

from ._api_compat import compile
from ._api_compat import filter
from ._api_compat import get_args
from ._api_compat import global_args
from ._api_compat import input
from ._api_compat import merge_outputs
from ._api_compat import output
from ._api_compat import overwrite_output
from ._api_compat import run
from ._api_compat import run_async
from .filter_graph import Stream
from .filters import *  # noqa: F403 this is impossible to avoid
from .generated_filters import *  # noqa: F403 this is impossible to avoid
from .probe import probe
from .view import view

generated_filters: Any
_api_compat: Any

# This is for `from xyz import *`, but also to make linter shut up

__all__ = [
    "input",
    "filter",
    "output",
    "global_args",
    "get_args",
    "compile",
    "run",
    "run_async",
    "probe",
    "view",
    "generated_filters",
    "overwrite_output",
    "merge_outputs",
]

stream_modules = [generated_filters, _api_compat]  # noqa: F821

for module in stream_modules:
    for name, func in getmembers(module, isfunction):
        setattr(Stream, name, func)



import subprocess


def probe(filename: str, format="json", **kwargs) -> str:
    """
    Calls ffprobe on given filename and returns output as string.

    :param filename: the graph to compile
    :param format: format of output stream. Available options: 'json', 'default', 'flat', 'csv', 'ini', 'xml' (default: json)
    :return: ffprobe output in given format
    """
    ffprobe_executable_path = "ffprobe"
    args = [ffprobe_executable_path, filename, "-of", format]
    args += ["-show_format", "-show_streams"]

    # parsing kwargs
    for key, value in kwargs.items():
        args.append(f"-{key}")
        if value is not None:
            args.append(str(value))

    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # noqa: S603
    out, err = p.communicate()

    if p.returncode != 0:
        raise Exception("ffprobe", out, err)

    return out.decode("utf-8")



import dataclasses
from enum import Enum
from re import sub
from typing import List
from typing import Optional

from .filter_graph import AnyNode
from .filter_graph import Filter
from .filter_graph import MergeOutputFilter
from .filter_graph import SinkFilter
from .filter_graph import SourceFilter
from .filter_graph import Stream


class NodeColors(Enum):
    INPUT = "#99cc00"
    OUTPUT = "#99ccff"
    FILTER = "#ffcc00"


@dataclasses.dataclass(eq=True, frozen=True)
class PrepNode:
    name: str
    color: NodeColors
    path: str

    def create_path_for_next(self) -> str:
        if "|" in self.path:
            return self.name
        sep = ""
        if self.path:
            sep = ";"
        return f"{self.path}{sep}{self.name}"

    def prev_node(self) -> List[str] | None:
        if not self.path:
            return None
        return [path.split(";")[-1] for path in self.path.split("|")]

    @property
    def id(self):
        if (string := sub(r"\w+\((\d+)\)", r"\1", self.name)) == self.name or "." in self.name:
            return 1
        else:
            return int(string)

    @property
    def command(self):
        if "." in self.name:
            return self.name
        else:
            return sub(r"(\w+)\(\d+\)", r"\1", self.name)


def create_graph_connections(parent_node: AnyNode | Stream, previous: List[PrepNode]) -> None:
    """Creates list of PrepNode objects that contain stream data to later
    transform it to a graph
    """
    new_connections = []
    nodes = None
    if isinstance(parent_node, Filter):
        nodes = parent_node._in
    else:
        nodes = parent_node._nodes
    for node in nodes:
        if isinstance(node, SourceFilter):
            new_connections.append(PrepNode(node.in_path.split("/")[-1], NodeColors.INPUT, ""))
        elif isinstance(node, SinkFilter):
            new_connections.append(PrepNode(node.out_path.split("/")[-1], NodeColors.OUTPUT, new_connections[-1].create_path_for_next()))
        elif isinstance(node, Filter):
            path = ""
            if not new_connections:
                create_graph_connections(node, previous)
                paths = []
                for stream in node._in:
                    last_node = stream._nodes[-1]
                    parent_prep_node = None
                    if isinstance(last_node, Filter):
                        parent_prep_node = next(
                            prep_node for prep_node in previous if ((prep_node.command, prep_node.id) == (last_node.command, last_node._id))
                        )
                    else:
                        parent_prep_node = next(prep_node for prep_node in previous if prep_node.name == last_node.in_path)
                    paths.append(parent_prep_node.create_path_for_next())
                path = "|".join(paths)
            else:
                path = new_connections[-1].create_path_for_next()
            suffix = "" if node._id < 2 else f"({node._id})"
            new_connections.append(PrepNode(f"{node.command}{suffix}", NodeColors.FILTER, path))
        elif isinstance(node, MergeOutputFilter):
            for stream in node.streams:
                create_graph_connections(stream, previous)
            return
        elif isinstance(node, Stream):
            create_graph_connections(node, previous)
    previous.extend(new_connections)


def view(graph: Stream, filename: Optional[str] = None) -> None:
    "Creates a graph of filters"

    import networkx as nx  # type: ignore
    from matplotlib import pyplot as plt  # type: ignore

    G = nx.DiGraph()

    graph_connection = []
    create_graph_connections(graph, graph_connection)
    graph_connection = list(dict.fromkeys(graph_connection))

    # Adding nodes
    for pre_node in graph_connection:
        G.add_node(pre_node.name, color=pre_node.color)

    # Adding edges
    for pre_node in graph_connection:
        if (prev := pre_node.prev_node()) is not None:
            for p in prev:
                G.add_edge(p, pre_node.name)

    pos = nx.circular_layout(G)
    nx.draw(
        G,
        pos,
        with_labels=True,
        node_shape="s",
        node_size=3000,
        node_color=[node.color.value for node in graph_connection],
        font_weight="bold",
    )

    if filename:
        plt.savefig(filename)
    else:
        plt.show()
